/*
 * Copyright (c) 2003 Orion Hodson <oho@acm.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "feeder_compat.h"
#include "feeder_random.c"
#include "feeder_stdio.c"
#include "feeder_infinite.c"
#include "feeder_finite.c"

#ifdef FEEDER_RATE_NEW
#include "feeder_rate_new.c"
#else
#include "feeder_rate_fbsd.c"
#endif

static void
sigint_handler(int sig)
{
	fprintf(stderr, "Caught SIGINT, exiting...\n");
	exit(sig);
}

static void
usage(void)
{
	fprintf(stderr, "feed_rate [ -c <channels> ] <src_rate> <dst_rate>\n");
	exit(EX_USAGE);
}

static void
probe(void)
{
	uint32_t src, dst;

	struct pcm_feeder *f;

	f = FEEDER_CREATE(feeder_rate);
	for (src = 4000; src < 48000; src += 25) {
		for (dst = 4000; dst < 48000; dst += 25) {
			if (FEEDER_SET(f, FEEDRATE_SRC, src) != 0) {
				fprintf(stderr, "Failed to set src rate to %d (dst = %d)\n",
					src, dst);
			}
			if (FEEDER_SET(f, FEEDRATE_DST, dst) != 0) {
				fprintf(stderr, "- Failed to set dst rate to %d (src = %d)\n", 
					dst, src);
			}
		}
	}
	FEEDER_DESTROY(f);
	exit(0);
}

int
main(int argc, char * const argv[])
{
	uint32_t srate, drate, channels;
	int dummy_io, random_fetch_sizes, max_fetch;
	int ch;

	struct pcm_feeder *in, *out;

	channels = 1;
	signal(SIGINT, sigint_handler);

	dummy_io = 0;
	random_fetch_sizes = 0;
	max_fetch = 0;
	
	while ((ch = getopt(argc, argv, "c:nm:pr")) != -1) {
		switch (ch) {
		case 'c':
			channels = atoi(optarg);
			break;
		case 'n':
			dummy_io = 1;
			break;
		case 'm':
			max_fetch = atoi(optarg);
			break;
		case 'p':
			probe();
			break;
		case 'r':
			random_fetch_sizes = 1;
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	    
	if (argc != 2)
		usage();

	srate = atoi(argv[0]);
	drate = atoi(argv[1]);

	if (channels != 2) {
		fprintf(stderr, 
			"Number of channels must be 2\n");
		usage();
	}

	if (srate > 96000 || srate == 0) {
		fprintf(stderr, 
			"Source sample rate must be in range (1-96000)\n");
		usage();
	}

	if (drate > 96000 || drate == 0) {
		fprintf(stderr, 
			"Destination sample rate must be in range (1-96000)\n");
		usage();
	}

	if (srate == 0 || drate == 0)
		usage();

	if (dummy_io) {
	    in = FEEDER_CREATE(feeder_nullin);
	} else {
	    in = FEEDER_CREATE(feeder_stdin);
	}
	if (NULL == in) {
	    fprintf(stderr, "Could not create input feeder.\n");
	    exit(-1);
	}

	if (max_fetch) {
		out = FEEDER_CREATE(feeder_finite);
		if (NULL == out) {
			fprintf(stderr, "Could not create finite feeder.\n");
			exit(-1);
		}
		FEEDER_SET(out, FEEDER_FINITE_MAXBYTES, max_fetch);
		in = FEEDER_JOIN(in, out);
	}
	
	if (random_fetch_sizes) {
	    out = FEEDER_CREATE(feeder_random);
	    if (NULL == out) {
		fprintf(stderr, "Could not create randomizer feeder.\n");
		exit(-1);
	    }
	    in = FEEDER_JOIN(in, out);
	}
	
	out = FEEDER_CREATE(feeder_rate);
	if (NULL == out) {
	    fprintf(stderr, "Could not create rate feeder.\n");
	    exit(-1);
	}
	in = FEEDER_JOIN(in, out);

	if (FEEDER_SET(out, FEEDRATE_SRC, srate) < 0) {
		fprintf(stderr, "Could not set source rate\n");
		exit(EX_DATAERR);
	}

	if (FEEDER_SET(out, FEEDRATE_DST, drate) < 0) {
		fprintf(stderr, "Could not set destination rate\n");
		exit(EX_DATAERR);
	}
	
	if (dummy_io) {
	    out = FEEDER_CREATE(feeder_nullout);
	} else {
	    out = FEEDER_CREATE(feeder_stdout);
	}
	if (NULL == out) {
	    fprintf(stderr, "Could not create stdout feeder.\n");
	    exit(-1);
	}
	in = FEEDER_JOIN(in, out);

	while (!FEEDER_EOF(out))
		FEEDER_FEED(out, NULL, NULL, 4096, NULL);

	while (out != NULL) {
	    in = out->source;
	    FEEDER_DESTROY(out);
	    out = in;
	}

	return 0;
}
