
#ifdef _KERNEL

#include <dev/sound/pcm/sound.h>
#include "feeder_if.h"
SND_DECLARE_FILE("$FreeBSD$");

#endif /* _KERNEL */

MALLOC_DEFINE(M_SINCFEEDER, "sincfeed", "pcm sinc feeder");

#include "feeder_sinc.h"

#define SINC_ASSERT(x, y...) 						      \
	do {								      \
    	if (!(x)) {							      \
	    printf y;							      \
	    abort();							      \
	    }								      \
	} while (0)

#ifdef SINC_DEBUG
#define SINC_TRACE(x...)  fprintf(stderr, x) 
#else
#define SINC_TRACE(x...)
#endif

struct feed_sinc_info;

typedef int (*sinc_filter_method)(struct feed_sinc_info*, uint32_t, int16_t*);

#define FEEDBUFSZ	8192
#define MAX_COEFF	(2 * SIGMA_MAX_WIDTH - 1)

#define GAIN		75
#define MAX_GAIN	100

#define MAX_Q		10

struct feed_sinc_info {
	int src;			/* src rate */
	int dst;			/* dst_rate */

	size_t	 n_coeff;		/* number of filter coefficients */
	int32_t  coeff[MAX_COEFF];	/* filter coefficients */

	int16_t  buffer[FEEDBUFSZ];	/* input buffer */
	uint16_t buffer_ticks;		/* number of available samples */
	uint16_t buffer_pos;		/* next available sample */
	uint8_t	 channels;		/* 1 = mono, 2 = stereo */

	uint8_t	 quality;		/* quality (0-MAX_Q) -> filter size */
};

#define BYTES_PER_SAMPLE	2
#define samples_per_tick(info)	(info->channels)
#define bytes_per_tick(info)	(info->channels * BYTES_PER_SAMPLE)
#define history_ticks(info)	(info->n_coeff)
#define history_samples(info)	(history_ticks(info) * samples_per_tick(info))
#define history_bytes(info)	(history_ticks(info) * bytes_per_tick(info))

#define max_buffer_bytes(info)	(FEEDBUFSZ - history_bytes(info))
#define max_buffer_ticks(info)	(max_buffer_bytes(info) / bytes_per_tick(info))

/* Clip operation - cheap in comparison to multiply adds */
#define clip16(x) if (x > 32767) x = 32767; else if (x < -32767) x = -32767;

static void
normalize(int32_t *sb, uint32_t n)
{
	uint32_t i;
	int32_t num, denom, total;

	/* Input samples are in range -32768--32767. */
	for (i = 0, total = 0; i < n; i++) {
		total += sb[i];
	}

	num = 32767;
	denom = total;

	for (i = 0; i < n; i++) {
		sb[i] = sb[i] * num / denom;
	}

	for (i = 0, total = 0; i < n; i++) {
		if (sb[i] < 0) {
			total += -sb[i];
		} else {
			total += sb[i];
		}
	}
}

static void
init_sinc_filter(struct feed_sinc_info* info, int sigma_index, int sinc_index)
{
	size_t mid_coeff, i;
	int32_t tmp;

	memset(info->coeff, 0, sizeof(info->coeff[0]) * MAX_COEFF);

	mid_coeff = sigma_tab[sigma_index].ncoeff - 1;
	info->n_coeff = 1 + 2 * mid_coeff;

	tmp = sigma_tab[sigma_index].coeff[0] * 
		sinc_tab[sinc_index].coeff[0];
	tmp >>= sigma_precision;
	info->coeff[mid_coeff] = tmp;

	/*
	 * The tabulated coefficients represent 1/2 of filter width, 
	 * calculate convoluted coefficients, scale, and mirror.
	 */
	for (i = 1; i <= mid_coeff; i++) {
		tmp = sigma_tab[sigma_index].coeff[i] * 
			sinc_tab[sinc_index].coeff[i];
		tmp >>= sigma_precision;
		info->coeff[mid_coeff + i] = tmp;
		info->coeff[mid_coeff - i] = tmp;
	}
	normalize(info->coeff, info->n_coeff);
	for (i = 0; i < info->n_coeff; i++) {
		info->coeff[i] *= GAIN;
		info->coeff[i] /= MAX_GAIN;
		info->coeff[i] >>= 2;
	}
#if 0
	for (i = 0; i < info->n_coeff; i++) {
		fprintf(stderr, "a[%d] = %d\n", i, info->coeff[i]);
	}
#endif
}

static int
feed_sinc_setup(struct pcm_feeder *f)
{
	struct feed_sinc_info *info = f->data;
	size_t sigma_index, sinc_index;

	if (info->src > info->dst) {
		sinc_index = sinc_table_index(info->dst, info->src);
	} else {
		sinc_index = sinc_table_index(info->src, info->dst);
	}

	if (sinc_index > 2)
		sinc_index -= 2;
	else
		sinc_index = 0;
	fprintf(stderr, "Sinc index = %d\n", sinc_index);

	sigma_index = info->quality * SIGMA_TABLE_SIZE / MAX_Q;
	if (sigma_index >= SIGMA_TABLE_SIZE)
		sigma_index = SIGMA_TABLE_SIZE - 1;

	init_sinc_filter(info, sigma_index, sinc_index);

	return 0;
}

static int
feed_sinc_init(struct pcm_feeder *f)
{
	struct feed_sinc_info *info;

	info = malloc(sizeof(*info), M_SINCFEEDER, M_NOWAIT | M_ZERO);
	if (info == NULL)
		return ENOMEM;

	info->src = 8000;
	info->dst = 8000;
	info->quality = 4;
	info->channels = 2;
	info->buffer_ticks = 0;
	info->buffer_pos = 0;

	f->data = info;

	return feed_sinc_setup(f);
}

static int
feed_sinc_free(struct pcm_feeder *f)
{
	struct feed_sinc_info *info = f->data;

	if (info) {
		free(info, M_SINCFEEDER);
		f->data = NULL;
	}
	return 0;
}

static int
feed_sinc_set(struct pcm_feeder *f, int what, int value)
{
	struct feed_sinc_info *info = f->data;

	switch(what) {
	case FEEDSINC_SRC:
		info->src = value;
		break;
	case FEEDSINC_DST:
		info->dst = value;
		break;
	default:
		return -1;
	}
	return feed_sinc_setup(f);
}

static int
feed_sinc_get(struct pcm_feeder *f, int what)
{
	struct feed_sinc_info *info = f->data;

	switch(what) {
	case FEEDSINC_SRC:
		return info->src;
	case FEEDSINC_DST:
		return info->dst;
	default:
		return -1;
	}
	return -1;
}

static int
sinc_filter(struct feed_sinc_info *info, uint32_t ticks, int16_t* dst)
{
	uint32_t s, es, i, delta, delta1, delta2, delta3;
	int16_t *src;
	int32_t tmp, x, m[4];
	int32_t xb[16];

	src = info->buffer + info->buffer_pos * samples_per_tick(info);
	es = ticks * samples_per_tick(info);
	/* This loop is unrolled - we do 16 ops in a round, this seems to
	 * be about the optimum limit on the p4 box being used for testing.
	 * 32 decreases performance slightly.  It doesn't really matter that
	 * much.  This would be ripe for SIMD manipulation.
	 */
	s = 0;
	while (es - s > 16) {
	    memset(xb, 0, sizeof(xb));
	    for (i = 0; info->n_coeff - i >= 4; i += 4) {
		delta = i * info->channels + s;
		delta1 = (i + 1) * info->channels + s;
		delta2 = (i + 2) * info->channels + s;
		delta3 = (i + 3) * info->channels + s;		
		m[0] = info->coeff[i];
		m[1] = info->coeff[i + 1];
		m[2] = info->coeff[i + 2];
		m[3] = info->coeff[i + 3];		
		xb[ 0] += src[ 0 + delta] * m[0] + src[ 0 + delta1] * m[1] + src[ 0 + delta2] * m[2] + src[ 0 + delta3] * m[3];
		xb[ 1] += src[ 1 + delta] * m[0] + src[ 1 + delta1] * m[1] + src[ 1 + delta2] * m[2] + src[ 1 + delta3] * m[3];
		xb[ 2] += src[ 2 + delta] * m[0] + src[ 2 + delta1] * m[1] + src[ 2 + delta2] * m[2] + src[ 2 + delta3] * m[3];
		xb[ 3] += src[ 3 + delta] * m[0] + src[ 3 + delta1] * m[1] + src[ 3 + delta2] * m[2] + src[ 3 + delta3] * m[3];
		xb[ 4] += src[ 4 + delta] * m[0] + src[ 4 + delta1] * m[1] + src[ 4 + delta2] * m[2] + src[ 4 + delta3] * m[3];
		xb[ 5] += src[ 5 + delta] * m[0] + src[ 5 + delta1] * m[1] + src[ 5 + delta2] * m[2] + src[ 5 + delta3] * m[3];
		xb[ 6] += src[ 6 + delta] * m[0] + src[ 6 + delta1] * m[1] + src[ 6 + delta2] * m[2] + src[ 6 + delta3] * m[3];
		xb[ 7] += src[ 7 + delta] * m[0] + src[ 7 + delta1] * m[1] + src[ 7 + delta2] * m[2] + src[ 7 + delta3] * m[3];
		xb[ 8] += src[ 8 + delta] * m[0] + src[ 8 + delta1] * m[1] + src[ 8 + delta2] * m[2] + src[ 8 + delta3] * m[3];
		xb[ 9] += src[ 9 + delta] * m[0] + src[ 9 + delta1] * m[1] + src[ 9 + delta2] * m[2] + src[ 9 + delta3] * m[3];
		xb[10] += src[10 + delta] * m[0] + src[10 + delta1] * m[1] + src[10 + delta2] * m[2] + src[10 + delta3] * m[3];
		xb[11] += src[11 + delta] * m[0] + src[11 + delta1] * m[1] + src[11 + delta2] * m[2] + src[11 + delta3] * m[3];
		xb[12] += src[12 + delta] * m[0] + src[12 + delta1] * m[1] + src[12 + delta2] * m[2] + src[12 + delta3] * m[3];
		xb[13] += src[13 + delta] * m[0] + src[13 + delta1] * m[1] + src[13 + delta2] * m[2] + src[13 + delta3] * m[3];
		xb[14] += src[14 + delta] * m[0] + src[14 + delta1] * m[1] + src[14 + delta2] * m[2] + src[14 + delta3] * m[3];
		xb[15] += src[15 + delta] * m[0] + src[15 + delta1] * m[1] + src[15 + delta2] * m[2] + src[15 + delta3] * m[3];
	    }

	    for (; i < info->n_coeff; i++) {
		delta = i * info->channels + s;
		m[0] = info->coeff[i];
		xb[ 0] += src[ 0 + delta] * m[0];
		xb[ 1] += src[ 1 + delta] * m[0];
		xb[ 2] += src[ 2 + delta] * m[0];
		xb[ 3] += src[ 3 + delta] * m[0];
		xb[ 4] += src[ 4 + delta] * m[0];
		xb[ 5] += src[ 5 + delta] * m[0];
		xb[ 6] += src[ 6 + delta] * m[0];
		xb[ 7] += src[ 7 + delta] * m[0];
		xb[ 8] += src[ 8 + delta] * m[0];
		xb[ 9] += src[ 9 + delta] * m[0];
		xb[10] += src[10 + delta] * m[0];
		xb[11] += src[11 + delta] * m[0];
		xb[12] += src[12 + delta] * m[0];
		xb[13] += src[13 + delta] * m[0];
		xb[14] += src[14 + delta] * m[0];
		xb[15] += src[15 + delta] * m[0];
	    }
#define pack16(n) xb[n] >>= 13; clip16(xb[n]); dst[s + n] = xb[n];
	    pack16(0);
	    pack16(1);
	    pack16(2);
	    pack16(3);
	    pack16(4);
	    pack16(5);	    
	    pack16(6);
	    pack16(7);
	    pack16(8);
	    pack16(9);
	    pack16(10);
	    pack16(11);
	    pack16(12);
	    pack16(13);
	    pack16(14);
	    pack16(15);	    
	    s += 16;
	}
	
	for (;s < es; s++) {
	    tmp = 0;
	    for (i = 0; i < info->n_coeff; i++) {
		x = src[s + i * info->channels] 
		    * info->coeff[i];
		tmp += x;
	    }
	    tmp >>= 13;
	    clip16(tmp);
	    dst[s] = tmp;
	}

	info->buffer_pos += ticks;
	SINC_ASSERT(info->buffer_pos <= info->buffer_ticks, 
		    ("Buffer overrun"));
	return ticks;
}

static int
feed_sinc(struct pcm_feeder	*f, 
	  struct pcm_channel	*c, 
	  uint8_t		*b,
	  uint32_t		 count, 
	  void			*source)
{
	struct feed_sinc_info *info = f->data;
	uint32_t done;
	int fetch, ticks;

	done = 0;
	while (done < count) {
		/* Slurp in more data if input buffer is not full */
		fetch = (max_buffer_ticks(info) - info->buffer_ticks) * bytes_per_tick(info);
		if (fetch > 0) {
			uint8_t *u8b;
			u8b = (uint8_t*)info->buffer;
			u8b += (info->buffer_ticks + history_ticks(info)) *
				bytes_per_tick(info);
			fetch = FEEDER_FEED(f->source, c, u8b, fetch, source);
			SINC_ASSERT((fetch % 4) == 0, ("uneven fetch"));
			if (fetch == 0 && 
			    info->buffer_pos == info->buffer_ticks)
				break;
			info->buffer_ticks += fetch / bytes_per_tick(info);
		}
		
		ticks = (count - done) / bytes_per_tick(info);
		if (ticks > (info->buffer_ticks - info->buffer_pos)) {
			ticks = info->buffer_ticks - info->buffer_pos;
		}
		
		done += sinc_filter(info, ticks, (int16_t*)(b + done))
			* bytes_per_tick(info);

		SINC_ASSERT(done <= count, ("Output buffer overrun"));

		/* Roll history full of data from end of buffer to start */
		if (info->buffer_pos == max_buffer_ticks(info)) {
			uint8_t *u8b;
			u8b = (uint8_t*)info->buffer;
			bcopy(u8b + max_buffer_bytes(info), u8b,
			      history_bytes(info));
			info->buffer_pos = 0;
			info->buffer_ticks = 0;
		}
	}
	return done;
}

static struct pcm_feederdesc feeder_sinc_desc[] = {
	{FEEDER_SINC, AFMT_S16_LE | AFMT_STEREO, AFMT_S16_LE | AFMT_STEREO, 0},
	{0, 0, 0, 0}
};

static kobj_method_t feeder_sinc_methods[] = {
    	KOBJMETHOD(feeder_init,		feed_sinc_init),
    	KOBJMETHOD(feeder_free,		feed_sinc_free),
    	KOBJMETHOD(feeder_set,		feed_sinc_set),
    	KOBJMETHOD(feeder_get,		feed_sinc_get),
    	KOBJMETHOD(feeder_feed,		feed_sinc),
	{ 0, 0 }
};

FEEDER_DECLARE(feeder_sinc, 2, NULL);

