/*
 * Copyright (c) 2003 Orion Hodson <oho@acm.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "feeder_compat.h"
#include "feeder_random.c"
#include "feeder_sinc.c"
#include "feeder_stdio.c"
#include "feeder_infinite.c"

static void
sigint_handler(int sig)
{
	fprintf(stderr, "Caught SIGINT, exiting...\n");
	exit(sig);
}

static void
usage(void)
{
	fprintf(stderr, "feed_sinc [ -c <channels> ] <src_rate> <dst_rate>\n");
	exit(EX_USAGE);
}

int
main(int argc, char * const argv[])
{
	uint32_t srate, drate, channels;
	int dummy_io, random_fetch_sizes = 0;
	int ch;

	struct pcm_feeder *in, *out;

	channels = 1;
	signal(SIGINT, sigint_handler);

	dummy_io = 0;
	random_fetch_sizes = 0;
	
	while ((ch = getopt(argc, argv, "c:nr")) != -1) {
		switch (ch) {
		case 'c':
			channels = atoi(optarg);
			break;
		case 'n':
			dummy_io = 1;
			break;
		case 'r':
			random_fetch_sizes = 1;
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	    
	if (argc != 2)
		usage();

	srate = atoi(argv[0]);
	drate = atoi(argv[1]);

	if (channels != 2) {
		fprintf(stderr, 
			"Number of channels must be 2\n");
		usage();
	}

	if (srate > 48000 || srate == 0) {
		fprintf(stderr, 
			"Source sample rate must be in range (1-48000)\n");
		usage();
	}

	if (drate > 48000 || drate == 0) {
		fprintf(stderr, 
			"Destination sample rate must be in range (1-48000)\n");
		usage();
	}

	if (srate == 0 || drate == 0)
		usage();

	if (dummy_io) {
	    in = FEEDER_CREATE(feeder_nullin);
	} else {
	    in = FEEDER_CREATE(feeder_stdin);
	}
	if (NULL == in) {
	    fprintf(stderr, "Could not create input feeder.\n");
	    exit(-1);
	}

	if (random_fetch_sizes) {
	    out = FEEDER_CREATE(feeder_random);
	    if (NULL == out) {
		fprintf(stderr, "Could not create randomizer feeder.\n");
		exit(-1);
	    }
	    in = FEEDER_JOIN(in, out);
	}
	
	out = FEEDER_CREATE(feeder_sinc);
	if (NULL == out) {
	    fprintf(stderr, "Could not create sinc feeder.\n");
	    exit(-1);
	}
	in = FEEDER_JOIN(in, out);

	if (FEEDER_SET(out, FEEDSINC_SRC, srate) < 0) {
		fprintf(stderr, "Could not set source rate\n");
		exit(EX_DATAERR);
	}

	if (FEEDER_SET(out, FEEDSINC_DST, drate) < 0) {
		fprintf(stderr, "Could not set destination rate\n");
		exit(EX_DATAERR);
	}
	
	if (dummy_io) {
	    out = FEEDER_CREATE(feeder_nullout);
	} else {
	    out = FEEDER_CREATE(feeder_stdout);
	}
	if (NULL == out) {
	    fprintf(stderr, "Could not create stdout feeder.\n");
	    exit(-1);
	}
	in = FEEDER_JOIN(in, out);

	while (FEEDER_FEED(out, NULL, NULL, 4096, NULL) != 0);

	while (out != NULL) {
	    in = out->source;
	    FEEDER_DESTROY(out);
	    out = in;
	}

	return 0;
}
