/*
 * Linear interpolation routines for sample rate conversion of mono and stereo
 * sound streams.
 *
 * Basic algorithm (taps = 2) uses linear interpolation and is implemented in
 * integer arithmetic.  Higher order (more taps) algoritms are available, but
 * only in mono and using floating point.  Informal listening tests indicate
 * little benefit of using more taps.
 *
 * $Id: linrate.c,v 1.8 2003/01/10 18:26:03 hodson Exp $
 *
 * (c) Orion Hodson, 2003.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>
#include <inttypes.h>

#define trace(x...) fprintf(stderr, x)

static uint32_t
gcd(uint32_t n1, uint32_t n2)
{
    return n2 ? gcd(n2, n1 % n2) : n1;
}

static int
log_base2(int n)
{
    int u = 0;
    while (n >> u) {
	u++;
    }
    return u;
}

struct interp {
	uint32_t srate;		/* source rate */
	uint32_t drate;		/* destination rate */
	uint32_t channels;	/* number of channels */

	uint32_t crate;		/* common rate */
	uint32_t sscale;	/* src clock scale */
	uint32_t dscale;	/* dst clock scale */

	uint32_t mscale;	/* scale factor to avoid divide per sample */
	uint32_t mroll;		/* roll to avoid divide per sample */
	
	int16_t l1;		/* last sample (left) */
	int16_t r1;		/* last sample (right) */
	int16_t l2;		/* next to last sample (left) */
	int16_t r2;		/* next to last sample (right) */
	int16_t l3;		/* next next to last sample (left) */
	int16_t r3;		/* next next to last sample (right) */

	uint32_t (*convert)(struct interp*, int16_t*, uint32_t, int16_t*, uint32_t);
};

uint32_t
convert_mono_2_tap(struct interp* interp,
		   int16_t* src, uint32_t src_samples,
		   int16_t* dst, uint32_t dst_samples);

uint32_t
convert_stereo_2_tap(struct interp* interp,
		     int16_t* src, uint32_t src_samples,
		     int16_t* dst, uint32_t dst_samples);

uint32_t
convert_mono_3_tap(struct interp* interp,
		   int16_t* src, uint32_t src_samples,
		   int16_t* dst, uint32_t dst_samples);

uint32_t
convert_mono_4_tap(struct interp* interp,
		   int16_t* src, uint32_t src_samples,
		   int16_t* dst, uint32_t dst_samples);


struct ConverterProperty {
	uint16_t taps;
	uint16_t channels;
	uint32_t (*convert)(struct interp*, int16_t*, uint32_t, int16_t*, uint32_t);
} converter[] = {
	{ 2, 1, convert_mono_2_tap },
	{ 2, 2, convert_stereo_2_tap },
	{ 3, 1, convert_mono_3_tap },
	{ 4, 1, convert_mono_4_tap },
};
#define ARRAY_END(x) (sizeof(x) / sizeof(x[0]))

static uint32_t
src_samples_per_cycle(const struct interp* interp) 
{
	return interp->crate * interp->channels / interp->sscale;
}

static uint32_t
dst_samples_per_cycle(const struct interp* interp) 
{
	return interp->crate * interp->channels / interp->dscale;
}

static int
init_mix_coeff(struct interp* interp)
{
	uint32_t mscale, mroll, l, r, u;
	for (l = 99; l > 90; l -= 3) {
		for (mscale = 1; mscale < 4096; mscale++) {
			u = interp->sscale * mscale;
			if (u > 0x3fff)
				break;
			mroll = log_base2(u);
			r = (u * 100) >> mroll;
			if (r > l && r <= 100) {
				interp->mscale = mscale;
				interp->mroll = mroll;
				trace("Converting %d to %d with mscale = %d "
				      "and mroll = %d (gain = %d / 100)\n",
				      interp->srate, interp->drate,
				      interp->mscale, interp->mroll, r);
				return 0;
			}
		}
		trace("Converting %d to %d failed to find mscale and mroll "
		      "(tolerance %d--100)\n", 
		      interp->srate, interp->drate, l);
	}
	return -1;
}

static int
init_interp(struct interp *interp, 
	    uint32_t srate, 
	    uint32_t drate, 
	    uint32_t channels,
	    uint16_t taps) 
{
	uint32_t g = gcd(srate, drate);

	interp->srate = srate;
	interp->drate = drate;
	interp->channels = channels;

	srate /= g;
	drate /= g;

	interp->crate = srate * drate;
	interp->sscale = drate;
	interp->dscale = srate;
	interp->l1 = 0;
	interp->r1 = 0;
	interp->l2 = 0;
	interp->r2 = 0;
	interp->l3 = 0;
	interp->r3 = 0;
	
	/* Enforce a minimum number of samples per calculation round */
	while (src_samples_per_cycle(interp) < 128 * interp->channels) {
		interp->crate *= 4;
		interp->sscale *= 2;
		interp->dscale *= 2;
	}

	interp->convert = 0;
	for (g = 0; g < ARRAY_END(converter); g++) {
		if (converter[g].taps == taps && 
		    converter[g].channels == channels) {
			interp->convert = converter[g].convert;
			break;
		}
	}
	
	if (interp->convert == 0) {
		fprintf(stderr, 
			"Unsupported channel and taps configuration\n");
		return -1;
	}
	
	return init_mix_coeff(interp);
}

/*****************************************************************************/
/* 2 tap integer arithmetic */

uint32_t
convert_mono_2_tap(struct interp *interp,
		   int16_t *src, uint32_t src_samples,
		   int16_t *dst, uint32_t dst_samples)
{
	uint32_t di, si, alpha, dalpha;

	int16_t l1;
	int32_t x;

	assert(src_samples == src_samples_per_cycle(interp));
	assert(dst_samples == dst_samples_per_cycle(interp));

	trace("sscale %d dscale %d crate %d\n",
	      interp->sscale, interp->dscale, interp->crate);

	dalpha = interp->dscale;

	trace("dalpha = %d\n", dalpha);

	alpha = 0;
	l1 = interp->l1;
	for (si = di = 0; di < dst_samples; di++) {
		assert(alpha <= interp->sscale);

		x = alpha * src[si] + (interp->sscale - alpha) * l1;

		{
			int32_t u;
			u = abs(x) * interp->mscale;
			u >>= interp->mroll;
			assert (u < 32768);
		}

		dst[di] = (interp->mscale * x) >> interp->mroll;

		alpha += dalpha;
		while (alpha >= interp->sscale) {
			alpha -= interp->sscale;
			l1 = src[si];
			si += 1;
		}
	}
	interp->l1 = l1;
	assert(si <= src_samples_per_cycle(interp));
	assert(di == dst_samples);
	return 0;
}

uint32_t
convert_stereo_2_tap(struct interp *interp,
		     int16_t *src, uint32_t src_samples,
		     int16_t *dst, uint32_t dst_samples)
{
	uint32_t di, si, alpha, dalpha;

	int16_t l1, r1;
	int32_t x;

	assert(src_samples == src_samples_per_cycle(interp));
	assert(dst_samples == dst_samples_per_cycle(interp));

	trace("sscale %d dscale %d crate %d\n",
	      interp->sscale, interp->dscale, interp->crate);

	dalpha = interp->dscale;

	trace("dalpha = %d\n", dalpha);

	alpha = 0;
	l1 = interp->l1;
	r1 = interp->r1;
	for (si = di = 0; di < dst_samples; di += 2) {
		assert(alpha < interp->crate);

		x = alpha * src[si] + (interp->sscale - alpha) * l1;
		dst[di] = (interp->mscale * x) >> interp->mroll;

		x = alpha * src[si + 1] + (interp->sscale - alpha) * r1;
		dst[di + 1] = (interp->mscale * x) >> interp->mroll;

		alpha += dalpha;
		if (alpha >= interp->sscale) {
			alpha -= interp->sscale;
			l1 = src[si];
			r1 = src[si + 1];
			si += 2;
		}
	}
	interp->l1 = l1;
	interp->r1 = r1;
	assert(si <= src_samples_per_cycle(interp));
	return 0;
}

/*****************************************************************************/
/* 3 tap floating point arithmetic */

uint32_t
convert_mono_3_tap(struct interp *interp,
		   int16_t *src, uint32_t src_samples,
		   int16_t *dst, uint32_t dst_samples)
{
	uint32_t di, si, alpha, dalpha;
	int16_t l1, l2;
	int32_t x;
	double fa, smpl;

	assert(src_samples == src_samples_per_cycle(interp));
	assert(dst_samples == dst_samples_per_cycle(interp));

	trace("sscale %d dscale %d crate %d\n",
	      interp->sscale, interp->dscale, interp->crate);

	dalpha = interp->dscale;

	trace("dalpha = %d\n", dalpha);

	alpha = 0;
	l1 = interp->l1;
	l2 = interp->l2;
	for (si = di = 0; di < dst_samples; di++) {
		assert(alpha <= interp->sscale);

		fa = (double)alpha / (double)interp->sscale;
		smpl = 0.5 * (fa - 1.0) * (fa - 2.0) * l2 -
			(fa) * (fa - 2.0) * l1 +
			0.5 * (fa) * (fa - 1.0) * src[si];

		x = (int32_t) smpl;
		assert(abs(x) <= 32768);
		dst[di] = (int16_t)x;

		alpha += dalpha;
		while (alpha >= interp->sscale) {
			alpha -= interp->sscale;
			l2 = l1;
			l1 = src[si];
			si += 1;
		}
	}
	interp->l1 = l1;
	interp->l2 = l2;

	assert(si <= src_samples_per_cycle(interp));
	assert(di == dst_samples);
	return 0;
}

/*****************************************************************************/
/* 4 tap floating point arithmetic */

uint32_t
convert_mono_4_tap(struct interp *interp,
		   int16_t *src, uint32_t src_samples,
		   int16_t *dst, uint32_t dst_samples)
{
	uint32_t di, si, alpha, dalpha;
	int16_t l1, l2, l3;
	int32_t x;
	double fa, smpl;

	assert(src_samples == src_samples_per_cycle(interp));
	assert(dst_samples == dst_samples_per_cycle(interp));

	trace("sscale %d dscale %d crate %d\n",
	      interp->sscale, interp->dscale, interp->crate);

	dalpha = interp->dscale;

	trace("dalpha = %d\n", dalpha);

	alpha = 0;
	l1 = interp->l1;
	l2 = interp->l2;
	l3 = interp->l3;
	for (si = di = 0; di < dst_samples; di++) {
		assert(alpha <= interp->sscale);

		fa = (double)alpha / (double)interp->sscale;
		smpl = - (fa - 1.0) * (fa - 2.0) * (fa - 3.0) * l3/6.0 +
			fa * (fa - 2.0) * (fa - 3.0) * l2 / 2.0 -
			fa * (fa - 1.0) * (fa - 3.0) * l1 / 2.0 +
			fa * (fa - 1.0) * (fa - 2.0) * src[si] / 6.0;
				
		x = (int32_t) smpl;
		assert(abs(x) <= 32768);
		dst[di] = (int16_t)x;

		alpha += dalpha;
		while (alpha >= interp->sscale) {
			alpha -= interp->sscale;
			l3 = l2;
			l2 = l1;
			l1 = src[si];
			si += 1;
		}
	}
	interp->l1 = l1;
	interp->l2 = l2;
	interp->l3 = l3;

	assert(si <= src_samples_per_cycle(interp));
	assert(di == dst_samples);
	return 0;
}

/*****************************************************************************/

static
void process_samples(struct interp *interp)
{
	int16_t sbuf[96000], dbuf[96000]; /* 2 channels @48kHz */

	uint32_t ssmpls, dsmpls, blocks;

	ssmpls = src_samples_per_cycle(interp);
	dsmpls = dst_samples_per_cycle(interp);

	blocks = 0;

	while (fread(sbuf, sizeof(sbuf[0]), ssmpls, stdin) == ssmpls) {
		interp->convert(interp, sbuf, ssmpls, dbuf, dsmpls);
		fwrite(dbuf, sizeof(dbuf[0]), dsmpls, stdout);
		fflush(stdout);
		blocks++;
	}
	fprintf(stderr, "Processed %d samples, generated %d\n",
		blocks * ssmpls, blocks * dsmpls);
}

static void
usage(void)
{
    fprintf(stderr, "linrate [ -c <channels> ] <src_rate> <dst_rate>\n");
    exit(EX_USAGE);
}

int 
main(int argc, char* const argv[])
{
	uint32_t srate, drate, channels, taps;
	int ch;
	struct interp interp;

	channels = 1;
	taps = 2;
    
	while ((ch = getopt(argc, argv, "c:t:")) != -1) {
		switch (ch) {
		case 'c':
			channels = atoi(optarg);
			break;
		case 't':
			taps = atoi(optarg);
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	    
	if (argc != 2)
		usage();

	srate = atoi(argv[0]);
	drate = atoi(argv[1]);

	if (channels != 1 && channels != 2) {
		fprintf(stderr, 
			"Number of channels must be in range (1-2)\n");
		usage();
	}

	if (srate > 48000 || srate == 0) {
		fprintf(stderr, 
			"Source sample rate must be in range (1-48000)\n");
		usage();
	}

	if (drate > 48000 || drate == 0) {
		fprintf(stderr, 
			"Destination sample rate must be in range (1-48000)\n");
		usage();
	}

	if (srate == 0 || drate == 0)
		usage();

	if (init_interp(&interp, srate, drate, channels, taps) < 0) {
		exit(-1);
	}

	process_samples(&interp);

	return 0;
}
