#include <sys/ioctl.h>

#include <machine/soundcard.h>

#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>

static void
usage(void)
{
        fprintf(stderr, 
		"rawrec [-b bits per sample] [ -c <channels> ] <smpl_rate>");
        exit(EX_USAGE);
}

static int
log_base2(int n)
{
    int u = 0;
    while (n >> u) {
	u++;
    }
    return u;
}

static void
record(const char*	sample_device, 
       uint32_t 	sample_rate, 
       uint32_t 	sample_channels, 
       uint32_t 	sample_encoding,
       int		blksz)
{

	snd_chan_param	scp;
	int		fd, frag;
	char		buffer[1024];
	size_t		buffer_bytes = sizeof(buffer) / sizeof(buffer[0]);

	fd = open(sample_device, O_RDONLY);
	if (fd < 0) {
		errx(1, "Could not open %s", sample_device);
	}

	if (sample_channels >= 2)
		sample_encoding |= AFMT_STEREO;

	scp.play_rate	= 0;
	scp.play_format	= 0;
	scp.rec_rate	= sample_rate;
	scp.rec_format	= sample_encoding;

	fprintf(stderr, "format = 0x%08x rate = %lu\n",
		(int)scp.rec_format, scp.rec_rate);
	if (ioctl(fd, AIOSFMT, &scp) < 0) {
		errx(1, "Could not set format and rate (0x%08x and %d Hz)",
		     sample_encoding, sample_rate);
	}

	/* Check set format */
	memset(&scp, 0, sizeof(scp));
	if (ioctl(fd, AIOGFMT, &scp) < 0) {
		errx(1, "Could not read device format");
	}
/*	assert(scp.play_rate == scp.play_format == 0);  */
	assert(scp.rec_rate == sample_rate);
	assert(scp.rec_format == sample_encoding);

	if (blksz > 0) {
		frag = log_base2(blksz) | 0xffff0000;
	
		if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag) < 0) {
			perror("SNDCTL_DSP_SETFRAGMENT");
			exit(EX_DATAERR);
		}
	}

	fprintf(stderr, "pre-read");
	read(fd, buffer, 4);
	fprintf(stderr, ".\n");

	for(;;) {
		ssize_t rd = read(fd, buffer, buffer_bytes);
		if (rd <= 0) {
			perror("read");
			fprintf(stderr, "read error.");
			break;
		}
		if (fwrite(buffer, 1, rd, stdout) == 0) {
			perror("write");
			break;
		}
	}
}

int main(int argc, char * const argv[]) 
{
	const char* sample_device = "/dev/audio";
	uint32_t sample_rate = 8000;
	uint32_t sample_channels = 1;
	uint32_t sample_encoding = AFMT_S16_LE;
	uint32_t bits = 16;
	int blksz;

	int ch;

	blksz = 512;
	
	while ((ch = getopt(argc, argv, "b:c:f:B:")) != -1) {
		switch(ch) {
		case 'b':	bits = atoi(optarg); break;
		case 'c':	sample_channels = atoi(optarg); break;
		case 'f':	sample_device = optarg; break;
		case 'B':	blksz = atoi(optarg); break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1)
                usage();

	sample_rate = atoi(argv[0]);
	if (sample_rate < 8000 || sample_rate > 48000) {
		fprintf(stderr, 
			"rate must lie in range 8000-48000\n");
		usage();
	}

	switch (bits) {
	case 8:
		sample_encoding = AFMT_S8;
		break;
	case 16:
		sample_encoding = AFMT_S16_LE;
		break;
	default:
		fprintf(stderr, "bits per sample must be 8 or 16\n");
		usage();
	}

	record(sample_device, sample_rate, sample_channels, sample_encoding,
	       blksz);

	return 0;
}
