#include <assert.h>
#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#define OUT_SAMPLES 256

typedef int16_t(sample_gen)(uint16_t amp, 
			    uint16_t sig_rate,
			    uint16_t smpl_rate,
			    uint32_t t);

static void
mono_gen(uint16_t amp, uint16_t trate, uint16_t srate,
	 sample_gen generator, int16_t* buf, uint32_t buffer_bytes)
{
	static uint32_t p = 0;
	uint32_t i, e;
	e = buffer_bytes / 2;
	for (i = 0 ; i < e; i++) {
		buf[i] = generator(amp, trate, srate, i + p);
	}
	p += e;
}

static void
stereo_gen(uint16_t amp, uint16_t trate, uint16_t srate,
	   sample_gen generator, int16_t* buf, uint32_t buffer_bytes)
{
	static uint32_t p = 0;
	uint32_t i, e;
	e = buffer_bytes / 4;
	for (i = 0 ; i < e; i++) {
		buf[2 * i] = generator(amp, trate, srate, i + p);
		buf[2 * i + 1] = buf[2 * i];
	}
	p += e;
}

static void
generate_samples(int16_t amp, uint16_t trate, uint16_t srate,
		 uint16_t bits, uint16_t channels, sample_gen generator)
{
	int16_t sbuf[512];
	size_t wr = 0;
	do {
		if (channels == 1) {
			mono_gen(amp, trate, srate, generator,
				 sbuf, sizeof(sbuf)); 
		} else {
			assert(channels == 2);
			stereo_gen(amp, trate, srate, generator,
				   sbuf, sizeof(sbuf));
		}
		if (bits == 16) {
			wr = fwrite(sbuf, sizeof(sbuf[0]), 
				    sizeof(sbuf) / sizeof(sbuf[0]), stdout);
		} else {
			size_t i;
			int8_t *sb8;
			assert(bits == 8);
			sb8 = (int8_t*)sbuf;
			for (i = 0; i < sizeof(sbuf) / sizeof(sbuf[0]); i++) {
				sb8[i] = (sbuf[i] >> 8);
			}
			wr = fwrite(sb8, 1, i, stdout);
		}
	} while (wr != 0);
}


static int16_t
sin_gen(uint16_t amp, uint16_t sig_rate, uint16_t smpl_rate, uint32_t t)
{
	double tmp;
	tmp = amp * sin(2 * M_PI * t * sig_rate / (double)smpl_rate);
	return tmp;
}

static int16_t
conv_gen(uint16_t amp, uint16_t sig_rate, uint16_t smpl_rate, uint32_t t)
{
	double tmp;
	tmp = sin(2 * M_PI * t * sig_rate * 0.05 / (double)smpl_rate);
	tmp *= tmp;
	tmp *= amp * sin(2 * M_PI * t * sig_rate / (double)smpl_rate);
	return tmp;
}

static int16_t
sq_gen(uint16_t amp, uint16_t sig_rate, uint16_t smpl_rate, uint32_t t)
{
	int32_t r;
	int16_t step_length = smpl_rate / sig_rate;

	if (step_length == 0)
		step_length = 1;
	r = -amp + ((t / step_length) % 2) * amp * 2;
	return r;
}

static int16_t
saw_gen(uint16_t amp, uint16_t sig_rate, uint16_t smpl_rate, uint32_t t)
{
	double tmp;
	int16_t step_length = smpl_rate / sig_rate;
	if (step_length == 0)
		step_length = 1;
	tmp = - 0.5 * amp + 1.0 * (t % step_length) / 
		(0.5 * step_length) * amp;
	return tmp;
}

struct gen {
	const char* name;
	sample_gen* method;
} generators[] = {
	{ "si",		sin_gen	 },
	{ "conv",	conv_gen },
	{ "sq",		sq_gen 	 },
	{ "saw",	saw_gen	 },
};
#define ngenerators (sizeof(generators) / sizeof(generators[0]))

static int
name_to_generator(const char* name)
{
	uint32_t i;
	for (i = 0; i < ngenerators; i++) {
		if (0 == strncasecmp(name, generators[i].name, 2)) {
			return i;
		}
	}
	return -1;
}

static void
usage(void)
{
	fprintf(stderr, 
		"tonegen [ -c <channels> ] [ -a <amplitute> ] [ -s sq|sin ]"
		"<tone_rate> <smpl_rate>");
	exit(EX_USAGE);
}

int main(int argc, char * const * argv)
{
	int channels, amp, trate, srate, shape, ch, bits;

	shape = 0;
	channels = 1;
	amp = 32767;
	bits = 16;

	while ((ch = getopt(argc, argv, "a:b:c:s:")) != -1) {
                switch (ch) {
		case 'a':
			amp = atoi(optarg);
			break;
		case 'b':
			bits = atoi(optarg);
			break;
                case 'c':
                        channels = atoi(optarg);
                        break;
		case 's':
			shape = name_to_generator(optarg);
			if (shape < 0) {
				fprintf(stderr, "Shape %s not known\n", optarg);
				usage();
			}
			break;
                case '?':
                default:
                        usage();
                }
        }
	argc -= optind;
        argv += optind;
            
        if (argc != 2)
                usage();

	trate = atoi(argv[0]);
	srate = atoi(argv[1]);

	if (channels < 1 || channels > 2) {
		fprintf(stderr, "Channels must be in range 1--2\n");
		usage();
	}

	if (amp < 1 || amp > 32767) {
		fprintf(stderr, "Amplitude must be in range 1--32767\n");
		usage();
	}

	if (bits != 8 && bits != 16) {
		fprintf(stderr, "bits must be 8 or 16.");
		usage();
	}

	if (srate == 0 || srate > 48000) {
		fprintf(stderr, "smpl_rate must be in range 1--48000");
		usage();
	}

	if (trate == 0 || trate >= srate / 2) {
		fprintf(stderr, "tone_rate must be in range 1--smpl_rate / 2");
		usage();
	}

	generate_samples(amp, trate, srate, bits, channels, 
			 generators[shape].method);

	return 0;
}
